;
; ideserv.asm - X1541 pclink handler
;
; Written by
;  Kajtar Zsolt <soci@c64.rulez.org>
;
;    This file is part of IDEDOS the IDE64 disk operating system
;    See README for copyright notice.
;
;    This program is free software; you can redistribute it and/or modify
;    it under the terms of the GNU General Public License as published by
;    the Free Software Foundation; either version 2 of the License, or
;    (at your option) any later version.
;
;    This program is distributed in the hope that it will be useful,
;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;    GNU General Public License for more details.
;
;    You should have received a copy of the GNU General Public License
;    along with this program; if not, write to the Free Software
;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
;
;---------------------------------------
; Initiate communication (0 is returned)
;
;          1             1
;       __________________________
; C64  | ACCU | SECONDARY ADDRESS |
;      |______|___________________|
;
pclink_secaddr  = next_sector+3
pclink_timer	= current_path
pclink_crc	= current_path+2

pclink_init	.proc
		pha
                #stz_a pclink_timer+1
                .if PCLINK_USECRC=1
                #stz_aa pclink_crc
                .fi
		
		.if PCLINK_TYPE=1
		lda #%11000111
;                sta pclink_timer+1
		and $dd00
		sta $dd00

lp		lda #63-8-32
		sta $dd02

		lda #63-32
		sta $dd02

                jsr pclink_timeout

                bit $dd00
                bvc lp

		lda #63
		sta $dd02
		.fi
		
		.if PCLINK_TYPE=2
		lda #$0f
		sta $dd03
		lda #63-4
		sta $dd02
		lda $dd00
		and #$fb
		sta $dd00
		.fi
		
                pla
		jsr pclink_send
                #stz_a pclink_timer+1
		lda pclink_secaddr
;---------------------------------------
; Send a byte
;
;     C64 ____RDY RDY_____RDY RDY_____RDY RDY_____RDY RDY____
; ATN         \_____/     \_____/     \_____/     \_____/
;
;     PC  __GO   GO_____GO   GO_____GO   GO_____GO   GO_____
; DATA      \_____/     \_____/     \_____/     \_____/     
;               |     |     |     |     |     |     |     |
;     DATA____ _____ _____ _____ _____ _____ _____ _____ ___
; CLK     ____X_7___X_6___X_5___X_4___X_3___X_2___X_1___X_0_
;
pclink_send     .proc

		.if PCLINK_USECRC=1
                ldy pclink_crc
                eor crc8,y
                sta pclink_crc
                eor crc8,y
                .fi

		.if PCLINK_TYPE=1
		sec
                rol
		bcs cs1

cs4		ldy #63-8
v2		bit $dd00
		bmi v1
		sty $dd02
                
                asl
                bcc cs2

cs3		ldy #63-16
v4		bit $dd00
		bpl v3
		sty $dd02

		asl
                bcc cs4
                beq cs5

cs1		ldy #63-8-16
v6		bit $dd00
		bmi v5
		sty $dd02

		asl
		bcs cs3
                
cs2		ldy #63
v8		bit $dd00
		bpl v7
		sty $dd02

		asl
                bcc cs4
                bne cs1
cs5		rts

v1		jsr pclink_timeout
                gne v2

v3		jsr pclink_timeout
                gne v4

v5		jsr pclink_timeout
                gne v6

v7		jsr pclink_timeout
                gne v8
		.fi
		
		.if PCLINK_TYPE=2
		sta $dd01
		lsr
		lsr
		lsr
		lsr
		tay
		lda #63
		sta $dd02
		lda #$10
v2		bit $dd0d
		beq v1
		sty $dd01
		ldy #63-4
		sty $dd02
v4		bit $dd0d
		beq v3
		rts

v1		jsr pclink_timeout
		gne v2

v3		jsr pclink_timeout
		gne v4
		.fi

		.pend
;---------------------------------------
; Receive a byte (Z and N flags must be set correctly on exit, Y not touched!)
; C=1 on timeout and exit caller
;
;     C64 _______OK   OK_____OK   OK_____OK   OK_____OK   OK____
; ATN            \_____/     \_____/     \_____/     \_____/
;
;     PC  _____RDY RDY_____RDY RDY_____RDY RDY_____RDY RDY______
; DATA         \_____/     \_____/     \_____/     \_____/     
;               |     |     |     |     |     |     |     |
;     DATA___ _____ _____ _____ _____ _____ _____ _____ ________
; CLK     ___X_7___X_6___X_5___X_4___X_3___X_2___X_1___X_0______
;         
pclink_recieve	.proc

		.if PCLINK_TYPE=1
                lda #$fe
lp      	
		
v2		ldx $dd00
		bmi v1
                cpx #$40
                
                ldx #63-8
		stx $dd02

		rol

v4              ldx $dd00
                bpl v3
                cpx #$c0
                
                ldx #63
                stx $dd02

                rol
                bcs lp

		.if PCLINK_USECRC=1
docrc           ldx pclink_crc
		eor crc8,x
                sta pclink_crc
                eor crc8,x
                .fi
		rts

v1		jsr pclink_timeout
		gne v2

v3		jsr pclink_timeout
                gne v4
		.fi
		
		.if PCLINK_TYPE=2

		lda #63
		sta $dd02
		lda #$10
v2		bit $dd0d
		beq v1
		ldx $dd01
		lda #63-4
		sta $dd02
		lda #$10
v4		bit $dd0d
		beq v3
		txa
		lsr
		lsr
		lsr
		lsr
		ora $dd01
                .if PCLINK_USECRC=1
                ldx pclink_crc
                eor crc8,x
                sta pclink_crc
                eor crc8,x
                .fi
		rts

v1		jsr pclink_timeout
		gne v2

v3		jsr pclink_timeout
                gne v4
		
		.fi
		.pend
;---------------------------------------
; Load a file
;
;         1            1                2        1   1        2        X     1
;       ___________________________________________                              ___
; C64  | "R" | SECONDARY ADDRESS | BYTES LO/HI | 0 |                            | CLOSE
;      |_____|___________________|_____________|___|                            |___
;   ___                                             ____________________________
; OPEN |                                           | ? | BYTES LO/HI| DATA | ST |
;   ___|                                           |___|____________|______|____|
;
;          1            1                2         1    1   1        2        X      1    1
;       __________________________________________________                                    ___
; C64  | !"R" | SECONDARY ADDRESS | BYTES LO/HI | CRC | 0 |                                  | CLOSE
;      |______|___________________|_____________|_____|___|                                  |___
;   ___                                                    __________________________________
; OPEN |                                                  | ? | BYTES LO/HI| DATA | CRC | ST |
;   ___|                                                  |___|____________|______|_____|____|
;
; $00 - OK
; $40 - EOF
; $42 - Too much
; 
pclink_load	.proc
		#xit_init
                bcs folyt

		lda #$60
                sta pclink_secaddr
		jsr pclink_send_open
		bne pclink_filenotf     ;readonly?

		.if PCLINK_USECRC=1
                lda #"R"
                .else
		lda #"r"
                .fi
                ldx #<$ffff
                .if SUPERCPU=1
                txy
                .else
                ldy #>$ffff
                .fi
                sec
                jsr pclink_dt
                bne kyp2

                jsr pclink_readsize

                inc next_sector
                bne cs1
                inc next_sector+1
                beq kyp2
cs1		jsr pclink_recieve
		tay
                inc next_sector
                bne cs2
                inc next_sector+1
                beq kyp2
cs2		jsr pclink_recieve
		ldx ActualSecAddr	;sec addr!=0?
		bne jklj
		ldy $c3
		lda $c4
jklj		sty $ae
		sta $af
                #lda0
		clc
                rts

folyt		ldy #0
eiqj		inc next_sector
                beq vx
vx2		jsr pclink_recieve
		jsr start_zeropage
		iny
		bne eiqj
                sty pclink_timer+1
                jsr pclink_stop
                lda $af
                cmp #$cf
                beq outm
		inc $af
		bne eiqj
outm            sec
                ldy #0
                rts			;out of memory

vx              inc next_sector+1
                bne vx2
		.if PCLINK_USECRC=1
                jsr pclink_recieve
                lda pclink_crc
                php
                .fi
		jsr pclink_recieve
                #phy_a
		jsr pclink_send_close
                #ply_a
		.if PCLINK_USECRC=1
                plp
                bne kyp2
                .fi
		clc
		rts			;exit

kyp2		lda #$02		;loaderr
		sta Status
                clc
                rts

pclink_filenotf	ldy #$42		;file not found
                sty Status
sto		sec
                rts
		.pend

peek_r		.proc
		.logical start_zeropage
		sta cfg_stnd		;3
at1		lda $ff00,x		;3
		.if SUPERCPU=1
                jsr switch_16k		;SCPU bug!
                .else
		sta cfg_open		;3
                .fi
		jmp pclink_send         ;3
		.here
		.pend

pclink_saveerr  .proc
		lda #$80
                sta Status
                rts
		.pend
                
;---------------------------------------
; Timeout handler
;
pclink_timeout	.proc
		dec pclink_timer
		bne vvx
		dec pclink_timer+1
		beq pclink_errexit
vvx             rts
		.pend
;---------------------------------------
; Stop handler
;
pclink_stop	.proc
		jsr check_stop
                bne pclink_errexit.mv

pclink_errexit 	.proc
		#xit_exit
		.if PCLINK_TYPE=1
		lda #63
		sta $dd02
		.fi
		
		.if PCLINK_TYPE=2
		lda #63-4
		sta $dd02
		.fi

		ldy #1			;break
                sec
mv		rts
		.pend
;---------------------------------------
; Blockwrite
;
; $00 - OK
;
pclink_write	.proc
		#xit_init
                
                #ldx_l_a buffer_output_exp
		lda buffer_address_hi_exp,x
                sec
                sbc buffer_start,x
                bcc none
                tay
                lda buffer_address_low_exp,x
                tax
                inx
                bne *+3
                iny
                jsr pclink_chrinout.in
                bcs vs2

		jsr OP_buffer_start
                #sbc1c
		sta buffer_address_hi_exp,x
		lda #255
		sta buffer_address_low_exp,x

none        	
		ldx #11
iukq		lda peek_r,x
		sta start_zeropage,x
		dex
		bpl iukq
                lda pom9
                sta peek_r.at1+1
                lda pom10
                sta peek_r.at1+2
		
                #ldatax write_size
                ora write_size+1
                clc
                beq vs2			;nothing to send

		txa
                eor #$ff
                sta pom9
                lda write_size+1
                tay
                eor #$ff
                sta pom10
                .if PCLINK_USECRC=1
                lda #"W"
                .else
		lda #"w"
                .fi
                clc
                jsr pclink_dt
                cmp #1
                bcs vs2			;crc error?

                tax
                geq xd

vv              jsr start_zeropage
		inx
                bne xd
                stx pclink_timer+1
                inc peek_r.at1+2
xd              inc pom9
		bne vv
                inc pom10
                bne vv
                jsr pclink_send_crc_0_r
                cmp #1			
                bcs vs2                 ;crc error?
                sta write_size
                sta write_size+1	;all sent
vs2		rts
		.pend
;---------------------------------------
; Open a file
;
;         1             1              X       1   1
;       _________________________________________
; C64  | "O" | SECONDARY ADDRESS | FILENAME  | 0 |
;      |_____|___________________|___________|___|
;                                                 ___
;                                                | ? |
;                                                |___|
;
;          1            1              X       1     1    1   1
;       ____________________________________________________
; C64  | !"O" | SECONDARY ADDRESS | FILENAME  | 0 | CRC | 0 |
;      |______|___________________|___________|___|_____|___|
;                                                            ___
;                                                           | ? |
;                                                           |___|
; $00 readonly
; $02 read+write
; $04 writeonly
; $06 error
;
pclink_send_open .proc
		jsr pclink_status_reset
		.if PCLINK_USECRC=1
                lda #"O"
                .else
		lda #"o"
                .fi
		jsr pclink_init
		ldx #0
exq		lda path_exp,x
		.if PCLINK_USECRC=1
		beq pclink_send_a_crc_0_r
                .else
                beq pclink_send_0_r
                .fi
		jsr pclink_send
		inx
		gne exq

pclink_send_a_crc_0_r .proc
		.if PCLINK_USECRC=1
                jsr pclink_send
pclink_send_crc_0_r .proc
		ldy pclink_crc
                lda crc8,y
                .fi

pclink_send_a_0_r .proc
		jsr pclink_send
                .if PCLINK_USECRC=0
pclink_send_crc_0_r .proc
                .fi
pclink_send_0_r .proc
                #lda0
                jsr pclink_send
		#jmps pclink_recieve
		.pend
;---------------------------------------
; Close a file
;
;         1             1          1
;       _____________________________
; C64  | "C" | SECONDARY ADDRESS | 0 |
;      |_____|___________________|___|
;                                     ___
;                                    | ? |
;                                    |___|
;
;          1             1           1    1
;       ____________________________________
; C64  | !"C" | SECONDARY ADDRESS | CRC | 0 |
;      |______|___________________|_____|___|
;                                            ___
;                                           | ? |
;                                           |___|
pclink_send_close
		jsr pclink_status_reset

		.proc
                .if PCLINK_USECRC=1
                lda #"C"
                .else
		lda #"c"
                .fi
		jsr pclink_init
                #jmps pclink_send_crc_0_r
		.pend
;---------------------------------------
; Open a file
;
pclink_openclose
		.proc
		#xit_init
                
                lda ActualSecAddr
                sta pclink_secaddr
                
                bcs pclink_send_close

		jsr pclink_send_open
                #andlsr 6
                tay
                lda modes,y
                rts
modes		.byte INFO_BLK_READ,INFO_STATUS,INFO_BLK_WRITE,INFO_IGNORE
		.pend
;---------------------------------------
; Fill chrin buffer
;
pclink_status	.proc

		#xit_init

		.if PCLINK_USECRC=1
                lda #"R"
                .else
		lda #"r"
                .fi
                ldx #<64
                ldy #>64
                clc
                jsr pclink_dt
                bne pclink_chrinout.vs
                lda #<pclink_status_exp
                sta pom1
                lda #>pclink_status_exp
                sta pom2
                .if SUPERCPU=1
                ldy #1
                sty pom3
                dey
                .else
                ldy #0
                .fi
                
                jsr pclink_readsize
                
                lda next_sector
                cmp #<($ffff-64)
                lda next_sector+1
                sbc #>($ffff-64)
                bcc pclink_chrinout.vs
                gcs pclink_chrinout.vv
;---------------------------------------
; Chrin
;
;         1            1           1   1   1   1        2        X     1
;       _____________________________________ 
; C64  | "R" | SECONDARY ADDRESS | 0 | 8 | 0 |
;      |_____|___________________|___|___|___|
;                                             ____________________________
;                                            | ? | BYTES LO/HI| DATA | ST |
;                                            |___|____________|______|____|
;
;          1            1           1   1    1    1   1        2        X      1    1
;       ____________________________________________ 
; C64  | !"R" | SECONDARY ADDRESS | 0 | 8 | CRC | 0 |
;      |______|___________________|___|___|_____|___|
;                                                    __________________________________
;                                                   | ? | BYTES LO/HI| DATA | CRC | ST |
;                                                   |___|____________|______|_____|____|
;
pclink_chrinout	.proc
		.if SUPERCPU=0
                txa
                .fi
		#xit_init
		.if SUPERCPU=0
                tax
                .fi
                
in              jsr pclink_status_reset

		bcs pclink_chrout
		.if PCLINK_USECRC=1
                lda #"R"
                .else
		lda #"r"
                .fi
                ldx #<2048
                ldy #>2048
                clc
                jsr pclink_dt
                bne vs

		jsr OP_buffer_start_pom1_y
                jsr pclink_readsize

vv              inc next_sector
		beq vx
vx2		jsr pclink_recieve
                #sta_zp_y pom1
                iny
                bne vv
                sty pclink_timer+1
                inc pom2
                lda pom2
                and #7
                bne vv
                geq ss
vx		inc next_sector+1
		bne vx2
ss              .if PCLINK_USECRC=1
		jsr pclink_recieve
                lda pclink_crc
                cmp #1
                php
		.fi
		jsr pclink_recieve
                .if PCLINK_USECRC=1
                plp
                .else
                clc
                .fi
vs		rts
		.pend
;---------------------------------------
; Flush chrout buffer
;
;         1            1                2        1   1    X     1   1
;       ___________________________________________     __________ 
; C64  | "W" | SECONDARY ADDRESS | BYTES LO/HI | 0 |   | DATA | 0 |
;      |_____|___________________|_____________|___|   |______|___|
;                                                   ___            ___
;                                                  | ? |          | ? |
;                                                  |___|          |___|
;
;         1             1                2         1    1   1    X      1    1   1
;       __________________________________________________     ________________
; C64  | !"W" | SECONDARY ADDRESS | BYTES LO/HI | CRC | 0 |   | DATA | CRC | 0 |
;      |______|___________________|_____________|_____|___|   |______|_____|___|
;                                                          ___                  ___
;                                                         | ? |                | ? |
;                                                         |___|                |___|
;
pclink_chrout	.proc
		txa
                eor #$ff
                sta pom4
                tya
                eor #$ff
                sta pom5
                .if PCLINK_USECRC=1
                lda #"W"
                .else
		lda #"w"
                .fi
                clc
                jsr pclink_dt
                bne vs

		jsr OP_buffer_start_pom1_y
                
                ldx pom4
                #jmps xd

vv              .if SUPERCPU=1
		lda [pom1]
                .else
		ldy #0
		lda (pom1),y
                .fi
                jsr pclink_send
                inc pom1
                bne *+4
                inc pom2
xd              inx
		bne vv
                stx pclink_timer+1
                inc pom5
                bne vv
                geq pclink_send_crc_0_r
vs		rts
		.pend
;---------------------------------------
; Data transfer start
;
;        1              1                2        1
;       ____________________________________________
; C64  | ACCU | SECONDARY ADDRESS | BYTES LO/HI | 0 |
;      |______|___________________|_____________|___|
;                                                    ___ 
;                                                   | ? |
;                                                   |___|
;
;         1              1                2         1    1
;       ___________________________________________________
; C64  | !ACCU | SECONDARY ADDRESS | BYTES LO/HI | CRC | 0 |
;      |_______|___________________|_____________|_____|___|
;                                                           ___
;                                                          | ? |
;                                                          |___|
pclink_dt	.proc
                sta pclink_timer+1
                #phy_a
                #phx_a
                bcs ajj
                #ldx_l_a buffer_currentp_exp
                lda buffer_pclink_secaddr,x
                sta pclink_secaddr
ajj             lda pclink_timer+1
                jsr pclink_init
                pla
                jsr pclink_send
                pla
                #jmps pclink_send_a_crc_0_r
                .pend
;---------------------------------------
; Status reset
;
pclink_status_reset .proc
                lda #$80
                sta status_pclink_pos_exp
		rts
                .pend
;---------------------------------------
; Read in size
;
;        2
;  ____________
; | BYTES LO/HI|
; |____________|

pclink_readsize .proc
                jsr pclink_recieve
                sta next_sector
                jsr pclink_recieve
                sta next_sector+1
                rts
		.pend
;---------------------------------------
; Blockread
;
		.comment
pclink_read	.proc
		#xit_init

                #ldatax buffer_currentp_exp
                lda buffer_pclink_secaddr,x
                sta pclink_secaddr

                lda #"r"
                jsr pclink_init
                lda #<2048
                jsr pclink_send
                lda #>2048
                jsr pclink_send
                jsr pclink_send
                jsr pclink_recieve
                cmp #0
                bne vs

		#ldatax buffer_currentp_exp
                lda buffer_start,x
                sta pom2
                .if SUPERCPU=1
                ldy #1
                sty pom3
                dey
                .else
                ldy #0
                .fi
                sty pom1

                jsr pclink_recieve
                sta next_sector
                jsr pclink_recieve
                sta next_sector+1

vv              inc next_sector
		beq vx
vx2		jsr pclink_recieve
                #sta_zp_y pom1
                iny
                bne vv
                inc pom2
                gne vv
vx		inc next_sector+1
		bne vx2
                jsr pclink_recieve
vs		rts
		.pend
                .endc
